// The deck state is stored as a string
// card,card,card|card,card,card,card,card|card,card,card
//      hand               deck               discarded


function shuffleArray(arr) {
    for (let i = arr.length - 1; i > 0; i--) {
        const j = Math.floor(Math.random() * (i + 1));
        [arr[i], arr[j]] = [arr[j], arr[i]]
    }
}

function getCardsInSection(sectionString) {
    let result = new Array()
    if (sectionString.length != 0) {
        let cards = sectionString.split(",")
        cards.forEach(e => {
            if (e.length != 0) {
                result.push(e)
            }
        })
    }

    return result
}

// returns three arrays [cards][deck][discard]
function getSections(deckString) {
    let result = new Array(3)
    let sections = deckString.split("|")
    for (let i = 0; i < 3; ++i) {
        result[i] = getCardsInSection(sections[i])
    }
    return result
}

function getDeckString(sections) {
    return sections[0].join(",") + "|" + sections[1].join(",") + "|" + sections[2].join(",")
}

// shuffles all cards (from hand, deck, and discard) and returns a new deck string
export function initShuffledDeck(deckString) {
    let sections = getSections(deckString)
    let allCards = sections[0].concat(sections[1], sections[2])

    shuffleArray(allCards)
    return getDeckString([[], allCards, []])
}

export function drawCards(deckString, numberToDraw) {
    let sections = getSections(deckString)
    for (let i = 0; i < numberToDraw; ++i) {
        if (sections[1].length == 0) {
            sections[1] = sections[2]
            sections[2] = []
            shuffleArray(sections[1])
        }
        sections[0].push(sections[1].shift())
    }
    return getDeckString(sections)
}

export function discardHand(deckString) {
    let sections = getSections(deckString)
    sections[2] = sections[2].concat(sections[0])
    sections[0] = []
    return getDeckString(sections)
}

export function getHand(deckString) {
    let sections = getSections(deckString)
    return sections[0].join(",")
}

export function handContains(deckString, card) {
    let sections = getSections(deckString)
    return sections[0].includes(card)
}